import torchvision
import os
import flgo.benchmark

transform = torchvision.transforms.Compose(
    [torchvision.transforms.ToTensor(),
     torchvision.transforms.Normalize((0.1307,), (0.3081,))]
)
path = os.path.join(flgo.benchmark.data_root, 'MNIST')
train_data = torchvision.datasets.MNIST(root=path, train=True, download=True, transform=transform)
test_data = torchvision.datasets.MNIST(root=path, train=False, download=True, transform=transform)



import os
import torchtext
import flgo.benchmark
import torch.nn
from torchtext.vocab import build_vocab_from_iterator
from torchtext.data.utils import get_tokenizer, ngrams_iterator

path = os.path.join(flgo.benchmark.data_root, 'SST2')
train_data = torchtext.datasets.SST2(root=path, split='train')
test_data = torchtext.datasets.SST2(root=path, split='dev')
ngrams = 2
tokenizer = get_tokenizer('basic_english')
def yield_tokens(data_iter, ngrams):
    for text,_ in data_iter:
        yield ngrams_iterator(tokenizer(text), ngrams)

vocab = build_vocab_from_iterator(yield_tokens(train_data, ngrams), specials=["<unk>"])
vocab.set_default_index(vocab["<unk>"])

def text_pipeline(x):
    return vocab(list(ngrams_iterator(tokenizer(x), ngrams)))

def label_pipeline(x):
    return int(x)

def apply_transform(x):
    return text_pipeline(x[0]), label_pipeline(x[1])

train_data = train_data.map(apply_transform)
test_data = test_data.map(apply_transform)

class TextClassificationModel(torch.nn.Module):
    def __init__(self, vocab_size, embed_dim, num_class):
        super(TextClassificationModel, self).__init__()
        self.embedding = torch.nn.EmbeddingBag(vocab_size, embed_dim, sparse=True)
        self.fc = torch.nn.Linear(embed_dim, num_class)
        self.init_weights()

    def init_weights(self):
        initrange = 0.5
        self.embedding.weight.data.uniform_(-initrange, initrange)
        self.fc.weight.data.uniform_(-initrange, initrange)
        self.fc.bias.data.zero_()

    def forward(self, text):
        offsets = [0]
        for t in text:
            offsets.append(t.size(0))
        offsets = torch.tensor(offsets[:-1]).cumsum(dim=0)
        text = torch.cat(text)
        offsets = offsets.to(text.device)
        embedded = self.embedding(text, offsets)
        return self.fc(embedded)

def get_model():
    return TextClassificationModel(vocab_size=len(vocab), embed_dim=64, num_class=4)


from flgo.experiment.logger import BasicLogger
import numpy as np

class PFLLogger(BasicLogger):
    def log_once(self, *args, **kwargs):
        # local performance
        cvals = []
        ctests = []
        for c in self.clients:
            model = c.model if (hasattr(c, 'model') and c.model is not None) else self.server.model
            cvals.append(c.test(model, 'val'))
            ctests.append(c.test(model, 'test'))
        cval_dict = {}
        if len(cvals)>0:
            for met_name in cvals[0].keys():
                if met_name not in cval_dict.keys(): cval_dict[met_name] = []
                for cid in range(len(cvals)):
                    cval_dict[met_name].append(cvals[cid][met_name])
                self.output['val_'+met_name].append(float(np.array(cval_dict[met_name]).mean()))
        ctest_dict = {}
        if len(ctests)>0:
            for met_name in ctests[0].keys():
                if met_name not in ctest_dict.keys(): ctest_dict[met_name] = []
                for cid in range(len(ctests)):
                    ctest_dict[met_name].append(ctests[cid][met_name])
                self.output['local_test_'+met_name].append(float(np.array(ctest_dict[met_name]).mean()))
        # global performance
        # gmetrics = self.server.test(self.server.model, 'test')
        # for met_name, met_val in gmetrics.items():
        #     self.output['global_test_' + met_name].append(met_val)
        # output to stdout
        self.show_current_output()


import os

import numpy as np
import torch.utils.data
from flgo.benchmark.base import BasicTaskCalculator
import flgo.benchmark.toolkits.cv.segmentation
from flgo.benchmark.toolkits.cv.classification import BuiltinClassPipe as ClsPipe
from flgo.benchmark.toolkits.cv.classification import BuiltinClassGenerator
from flgo.benchmark.toolkits.cv.detection.utils import mean_average_precision, precision_recall_ap, accuracy, redundancy, iou, average_precision
from sklearn.metrics import auc
import flgo.benchmark.base
import collections
try:
    import ujson as json
except:
    import json
from tqdm import tqdm

FromDatasetGenerator = flgo.benchmark.base.FromDatasetGenerator

FromDatasetPipe = flgo.benchmark.toolkits.cv.segmentation.FromDatasetPipe

class BuiltinClassPipe(ClsPipe):
    class TaskDataset(torch.utils.data.Subset):
        def __init__(self, dataset, indices, perturbation=None ):
            super().__init__(dataset, indices)
            self.dataset = dataset
            self.indices = indices
            self.perturbation = {idx: p for idx, p in zip(indices, perturbation)} if perturbation is not None else None

        def __getitem__(self, idx):
            if self.perturbation is None:
                if isinstance(idx, list):
                    return self.dataset[[self.indices[i] for i in idx]]
                return self.dataset[self.indices[idx]]
            else:
                return self.dataset[self.indices[idx]][0] + self.perturbation[self.indices[idx]], \
                       self.dataset[self.indices[idx]][1]

    def __init__(self, task_path, buildin_class, transform=None):
        super(BuiltinClassPipe, self).__init__(task_path, buildin_class, transform)

class GeneralCalculator(BasicTaskCalculator):
    r"""
    Calculator for the dataset in torchvision.datasets.

    Args:
        device (torch.device): device
        optimizer_name (str): the name of the optimizer
    """
    def __init__(self, device, optimizer_name='sgd'):
        super(GeneralCalculator, self).__init__(device, optimizer_name)
        self.criterion = torch.nn.CrossEntropyLoss()
        self.DataLoader = torch.utils.data.DataLoader
        self.collate_fn = lambda x:tuple(zip(*x))

    def compute_loss(self, model, data):
        """
        Args:
            model: the model to train
            data: the training dataset
        Returns: dict of train-one-step's result, which should at least contains the key 'loss'
        """
        model.train()
        tdata = self.to_device(data)
        output = model(*tdata)
        if hasattr(model, 'compute_loss'):
            loss = model.compute_loss(output, tdata)
        else:
            loss = sum(list(output.values()))
        return {'loss':loss}

    @torch.no_grad()
    def test(self, model, dataset, batch_size=64, num_workers=0, pin_memory=False):
        dataloader = self.get_dataloader(dataset, batch_size=batch_size, num_workers=num_workers, pin_memory=pin_memory, shuffle=False)
        num_classes = dataset.num_classes
        model.train()
        num_samples = 0
        losses = {}
        for batch_data in dataloader:
            batch_data = self.to_device(batch_data)
            output = model(*batch_data)
            if hasattr(model, 'compute_loss'):
                loss = model.compute_loss(output, batch_data).item()
            else:
                for k in output:
                    losses[k] = losses.get(k, 0.0) + output[k].item()*len(batch_data[0])
                loss = sum([v.item() for v in output.values()])
            losses['all_loss'] = losses.get('all_loss', 0.0) + loss*len(batch_data[0])
            num_samples += len(batch_data[0])
        for k,v in losses.items():
            losses[k]/=num_samples
        # compute AP
        predictions = []
        targets = []
        model.eval()
        for images, labels in tqdm(dataloader, desc='Predicting'):
            images = list(img.to(self.device) for img in images)
            labels = [{k: v.numpy() for k, v in t.items()} for t in labels]
            outputs = model(images)
            for out in outputs:
                for k in out.keys():
                    out[k] = out[k].cpu().numpy()
            predictions.extend(outputs)
            targets.extend(labels)
        # count TP for each class
        dects = {i:[] for i in range(1, num_classes)}
        gts = {i: {} for i in range(1, num_classes)}
        for image_id, pred in enumerate(predictions):
            for det_id in range(len(pred['boxes'])):
                class_id = int(pred['labels'][det_id])
                dects[class_id].append([image_id, class_id, pred['scores'][det_id], pred['boxes'][det_id]])
        for image_id, target in enumerate(targets):
            for gt_id in range(len(target['boxes'])):
                class_id = int(target['labels'][gt_id])
                gts[class_id][image_id] = gts[class_id].get(image_id, []) + [[image_id, class_id, [], target['boxes'][gt_id]]]
        res = []
        ious = np.arange(0.5, 1.0, 0.05)
        tf_dicts = {class_id:{iou_th:{'tp':None, 'fp':None} for iou_th in ious} for class_id in range(1, num_classes)}
        for class_id in range(1, num_classes):
            c_dects = sorted(dects[class_id], key=lambda d:d[2], reverse=True)
            c_gts = gts[class_id]
            c_tf_dict = tf_dicts[class_id]
            for iou_th in c_tf_dict:
                c_tf_dict[iou_th]['tp'] = np.zeros(len(c_dects))
                c_tf_dict[iou_th]['fp'] = np.zeros(len(c_dects))
            # c_tp = np.zeros(len(c_dects))
            # c_fp = np.zeros(len(c_dects))
            c_npos = sum(list(len(v) for v in c_gts.values()))
            for det_id in range(len(c_dects)):
                image_id = c_dects[det_id][0]
                gt = c_gts[image_id] if image_id in c_gts else []
                max_iou = -0.1
                for j in range(len(gt)):
                    d_iou = iou(gt[j][3], c_dects[det_id][3])
                    if d_iou> max_iou:
                        max_iou = d_iou
                        jmax = j
                for iou_th in ious:
                    if max_iou>iou_th:
                        if iou_th not in c_gts[c_dects[det_id][0]][jmax][2]:
                            c_gts[c_dects[det_id][0]][jmax][2].append(iou_th)
                            c_tf_dict[iou_th]['tp'][det_id] = 1
                            # c_tp[det_id] = 1
                        else:
                            c_tf_dict[iou_th]['fp'][det_id] = 1
                            # c_fp[det_id] = 1
                    else:
                        c_tf_dict[iou_th]['fp'][det_id] = 1
            res_ious = {}
            for iou_th in ious:
                c_acc_fp_i = np.cumsum(c_tf_dict[iou_th]['fp'])
                c_acc_tp_i = np.cumsum(c_tf_dict[iou_th]['tp'])
                c_recall_i = c_acc_tp_i/(c_npos+1e-8)
                c_precision_i = np.divide(c_acc_tp_i, (c_acc_tp_i + c_acc_fp_i))
                c_ap_i, c_mpre_i, c_mrec_i, c_ii_i = average_precision(c_recall_i, c_precision_i)
                res_ious[iou_th] = c_ap_i
            res.append(res_ious)
        # mAP@0.5
        tmp = [np.array([c_res[iou_th] for c_res in res]).mean() for iou_th in ious]
        mAP_05 = tmp[0]
        mAP_075 = tmp[5]
        mAP_05_095 = np.array(tmp).mean()
        ret = {}
        ret.update(losses)
        ret.update({'mAP@.5':float(mAP_05), 'mAP@.75':float(mAP_075), 'mAP@.5:.95':mAP_05_095})
        return ret

    def to_device(self, data):
        images, targets = data
        images = list(img.to(self.device) for img in images)
        targets = [{k: v.to(self.device) for k, v in t.items()} for t in targets]
        return images, targets

    def get_dataloader(self, dataset, batch_size=64, shuffle=True, num_workers=0, pin_memory=False, drop_last=False, *args, **kwargs):
        if self.DataLoader == None:
            raise NotImplementedError("DataLoader Not Found.")
        return self.DataLoader(dataset, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers, pin_memory=pin_memory, drop_last=drop_last, collate_fn=self.collate_fn)




"""This is a non-official implementation of 'Asynchronous Federated Optimization' (http://arxiv.org/abs/1903.03934). """
from flgo.algorithm.asyncbase import AsyncServer
from flgo.algorithm.fedprox import Client

class Server(AsyncServer):
    def initialize(self):
        self.init_algo_para(
            {'alpha': 0.6, 'mu': 0.005, 'flag': 'poly', 'hinge_a': 10, 'hinge_b': 6, 'poly_a': 0.5})

    def package_handler(self, received_packages:dict):
        if self.is_package_empty(received_packages): return False
        received_models = received_packages['model']
        taus = [m._round for m in received_models]
        alpha_ts = [self.alpha * self.s(self.current_round - tau) for tau in taus]
        currently_updated_models = [(1 - alpha_t) * self.model + alpha_t * model_k for alpha_t, model_k in zip(alpha_ts, received_models)]
        self.model = self.aggregate(currently_updated_models)
        return True

    def s(self, delta_tau):
        if self.flag == 'constant':
            return 1
        elif self.flag == 'hinge':
            return 1 if delta_tau <= self.hinge_b else 1.0 / (self.hinge_a * (delta_tau - self.hinge_b))
        elif self.flag == 'poly':
            return (delta_tau + 1) ** (-self.poly_a)



from torch import nn
import torch.nn.functional as F
from flgo.utils.fmodule import FModule

class Model(FModule):
    def __init__(self):
        super(Model, self).__init__()
        self.fc1 = nn.Linear(784, 200)
        self.fc2 = nn.Linear(200, 62)

    def forward(self, x):
        x = self.get_embedding(x)
        x = self.fc2(x)
        return x

    def get_embedding(self, x):
        x = x.view(-1, x.shape[1] * x.shape[-2] * x.shape[-1])
        x = self.fc1(x)
        x = F.relu(x)
        return x

def init_local_module(object):
    pass

def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.model = Model().to(object.device)


import torchvision
import os
import torch.nn as nn
import flgo.benchmark

transform = torchvision.transforms.Compose(
    [torchvision.transforms.ToTensor(),
     torchvision.transforms.Normalize((0.4914, 0.4822, 0.4465), (0.247, 0.243, 0.262))]
)

path = os.path.join(flgo.benchmark.data_root,  'CIFAR10')
train_data = torchvision.datasets.CIFAR10(root=path, train=True, download=True, transform=transform)
test_data = torchvision.datasets.CIFAR10(root=path, train=False, download=True, transform=transform)

class Model(nn.Module):
    def __init__(self):
        super(Model, self).__init__()
        self.embedder = nn.Sequential(
            nn.Conv2d(3, 64, 5),
            nn.ReLU(),
            nn.MaxPool2d(2),
            nn.Conv2d(64, 64, 5),
            nn.ReLU(),
            nn.MaxPool2d(2),
            nn.Flatten(1),
            nn.Linear(1600, 384),
            nn.ReLU(),
            nn.Linear(384, 192),
            nn.ReLU(),
        )
        self.fc = nn.Linear(192, 10)

    def forward(self, x):
        x = self.get_embedding(x)
        return self.fc(x)

    def get_embedding(self, x):
        return self.embedder(x)

def get_model():
    return Model()



r"""
This file contains preset partitioners for the benchmarks.
All the Partitioner should implement the method `__call__(self, data)`
where `data` is the dataset to be partitioned and the return is a list of the partitioned result.

For example, The IIDPartitioner.__call__ receives a indexable object (i.e. instance of torchvision.datasets.mnsit.MNSIT)
and I.I.D. selects samples' indices in the original dataset as each client's local_movielens_recommendation data.
The list of list of sample indices are finally returnerd (e.g. [[0,1,2,...,1008], ...,[25,23,98,...,997]]).

To use the partitioner, you can specify Partitioner in the configuration dict for `flgo.gen_task`.
 Example 1: passing the parameter of __init__ of the Partitioner through the dict `para`
>>>import flgo
>>>config = {'benchmark':{'name':'flgo.benchmark.mnist_classification'},
...            'partitioner':{'name':'IIDPartitioner', 'para':{'num_clients':20, 'alpha':1.0}}}
>>>flgo.gen_task(config, './test_partition')
"""
import warnings
from abc import abstractmethod, ABCMeta
import random

import networkx as nx
import numpy as np
import collections
import torch
from torch.utils.data import ConcatDataset
try:
    import community.community_louvain
except:
    pass

class AbstractPartitioner(metaclass=ABCMeta):
    @abstractmethod
    def __call__(self, *args, **kwargs):
        pass

class BasicPartitioner(AbstractPartitioner):
    """This is the basic class of data partitioner. The partitioner will be directly called by the
    task generator of different benchmarks. By overwriting __call__ method, different partitioners
    can be realized. The input of __call__ is usually a dataset.
    """
    def __call__(self, *args, **kwargs):
        return

    def register_generator(self, generator):
        r"""Register the generator as an self's attribute"""
        self.generator = generator

    def data_imbalance_generator(self, num_clients, datasize, imbalance=0, minvol=1):
        r"""
        Split the data size into several parts

        Args:
            num_clients (int): the number of clients
            datasize (int): the total data size
            imbalance (float): the degree of data imbalance across clients
            minvol (int): the minimal size of dataset
        Returns:
            a list of integer numbers that represents local data sizes
        """
        if imbalance == 0:
            samples_per_client = [int(datasize / num_clients) for _ in range(num_clients)]
            for _ in range(datasize % num_clients): samples_per_client[_] += 1
        else:
            imbalance = max(0.1, imbalance)
            sigma = imbalance
            mean_datasize = datasize / num_clients
            mu = np.log(mean_datasize) - sigma ** 2 / 2.0
            samples_per_client = np.random.lognormal(mu, sigma, (num_clients)).astype(int)
            crt_data_size = sum(samples_per_client)
            total_delta = np.abs(crt_data_size-datasize)
            thresold = max(int(total_delta/10), 1)
            delta = min(int(0.1 * thresold), 10)
            # force current data size to match the total data size
            while crt_data_size != datasize:
                if crt_data_size - datasize >= thresold:
                    maxid = np.argmax(samples_per_client)
                    maxvol = samples_per_client[maxid]
                    new_samples = np.random.lognormal(mu, sigma, (10 * num_clients))
                    while min(new_samples) > maxvol:
                        new_samples = np.random.lognormal(mu, sigma, (10 * num_clients))
                    new_size_id = np.argmin(
                        [np.abs(crt_data_size - samples_per_client[maxid] + s - datasize) for s in new_samples])
                    samples_per_client[maxid] = new_samples[new_size_id]
                elif crt_data_size - datasize >= delta:
                    maxid = np.argmax(samples_per_client)
                    if samples_per_client[maxid]>=delta:
                        samples_per_client[maxid] -= delta
                    elif samples_per_client[maxid]>1:
                        samples_per_client[maxid] -= 1
                elif crt_data_size - datasize > 0:
                    maxid = np.argmax(samples_per_client)
                    crt_delta = (crt_data_size - datasize)
                    if samples_per_client[maxid]>=crt_delta:
                        samples_per_client[maxid] -= crt_delta
                    elif samples_per_client[maxid]>=minvol:
                        samples_per_client[maxid] -= (crt_delta-minvol)
                    else:
                        warnings.warn("Failed to keep the minvol of clients' training data to be larger than {}".format(minvol))
                        if samples_per_client[maxid] > 1:
                            samples_per_client[maxid] -=1
                        else:
                            raise RuntimeError("Failed to generate distribution due to the conflicts of imbalance and num_clients. Please try to decrease the imbalance term or decrease the number of clients. ")
                elif datasize - crt_data_size >= thresold:
                    minid = np.argmin(samples_per_client)
                    minvol = samples_per_client[minid]
                    new_samples = np.random.lognormal(mu, sigma, (10 * num_clients))
                    while max(new_samples) < minvol:
                        new_samples = np.random.lognormal(mu, sigma, (10 * num_clients))
                    new_size_id = np.argmin(
                        [np.abs(crt_data_size - samples_per_client[minid] + s - datasize) for s in new_samples])
                    samples_per_client[minid] = new_samples[new_size_id]
                elif datasize - crt_data_size >= delta:
                    minid = np.argmin(samples_per_client)
                    samples_per_client[minid] += delta
                else:
                    minid = np.argmin(samples_per_client)
                    samples_per_client[minid] += (datasize - crt_data_size)
                crt_data_size = sum(samples_per_client)
            # let the minimal data size to be larger than 0
            while min(samples_per_client)==0:
                zero_client_idx = np.argmin(samples_per_client)
                maxid = np.argmax(samples_per_client)
                samples_per_client[maxid] -=1
                samples_per_client[zero_client_idx] += 1
            assert datasize==sum(samples_per_client) and min(samples_per_client)>0
        return samples_per_client

class IIDPartitioner(BasicPartitioner):
    """`Partition the indices of samples in the original dataset indentically and independently.

    Args:
        num_clients (int, optional): the number of clients
        imbalance (float, optional): the degree of imbalance of the amounts of different local_movielens_recommendation data (0<=imbalance<=1)
    """
    def __init__(self, num_clients=100, imbalance=0):
        self.num_clients = num_clients
        self.imbalance = imbalance

    def __str__(self):
        name = "iid"
        if self.imbalance > 0: name += '_imb{:.1f}'.format(self.imbalance)
        return name

    def __call__(self, data):
        samples_per_client = self.data_imbalance_generator(self.num_clients, len(data), self.imbalance)
        d_idxs = np.random.permutation(len(data))
        local_datas = np.split(d_idxs, np.cumsum(samples_per_client))[:-1]
        local_datas = [di.tolist() for di in local_datas]
        return local_datas

class DirichletPartitioner(BasicPartitioner):
    """`Partition the indices of samples in the original dataset according to Dirichlet distribution of the
    particular attribute. This way of partition is widely used by existing works in federated learning.

    Args:
        num_clients (int, optional): the number of clients
        alpha (float, optional): `alpha`(i.e. alpha>=0) in Dir(alpha*p) where p is the global distribution. The smaller alpha is, the higher heterogeneity the data is.
        imbalance (float, optional): the degree of imbalance of the amounts of different local_movielens_recommendation data (0<=imbalance<=1)
        error_bar (float, optional): the allowed error when the generated distribution mismatches the distirbution that is actually wanted, since there may be no solution for particular imbalance and alpha.
        index_func (func, optional): to index the distribution-dependent (i.e. label) attribute in each sample.
    """
    def __init__(self, num_clients=100, alpha=1.0, error_bar=1e-6, imbalance=0, index_func=lambda X:[xi[-1] for xi in X], minvol=1):
        self.num_clients = num_clients
        self.alpha = alpha
        self.imbalance = imbalance
        self.index_func = index_func
        self.minvol = minvol
        self.error_bar = error_bar

    def __str__(self):
        name = "dir{:.2f}_err{}".format(self.alpha, self.error_bar)
        if self.imbalance > 0: name += '_imb{:.1f}'.format(self.imbalance)
        return name

    def __call__(self, data):
        attrs = self.index_func(data)
        num_attrs = len(set(attrs))
        samples_per_client = self.data_imbalance_generator(self.num_clients, len(data), self.imbalance, minvol=self.minvol)
        # count the label distribution
        lb_counter = collections.Counter(attrs)
        lb_names = list(lb_counter.keys())
        p = np.array([1.0 * v / len(data) for v in lb_counter.values()])
        lb_dict = {}
        attrs = np.array(attrs)
        for lb in lb_names:
            lb_dict[lb] = np.where(attrs == lb)[0]
        proportions = [np.random.dirichlet(self.alpha * p) for _ in range(self.num_clients)]
        while np.any(np.isnan(proportions)):
            proportions = [np.random.dirichlet(self.alpha * p) for _ in range(self.num_clients)]
        sorted_cid_map = {k: i for k, i in zip(np.argsort(samples_per_client), [_ for _ in range(self.num_clients)])}
        error_increase_interval = 500
        max_error = self.error_bar
        loop_count = 0
        crt_id = 0
        crt_error = 100000
        while True:
            if loop_count >= error_increase_interval:
                loop_count = 0
                max_error = max_error * 10
            # generate dirichlet distribution till ||E(proportion) - P(D)||<=1e-5*self.num_classes
            mean_prop = np.sum([pi * di for pi, di in zip(proportions, samples_per_client)], axis=0)
            mean_prop = mean_prop / mean_prop.sum()
            error_norm = ((mean_prop - p) ** 2).sum()
            if crt_error - error_norm >= max_error:
                print("Error: {:.8f}".format(error_norm))
                crt_error = error_norm
            if error_norm <= max_error:
                break
            excid = sorted_cid_map[crt_id]
            crt_id = (crt_id + 1) % self.num_clients
            sup_prop = [np.random.dirichlet(self.alpha * p) for _ in range(self.num_clients)]
            del_prop = np.sum([pi * di for pi, di in zip(proportions, samples_per_client)], axis=0)
            del_prop -= samples_per_client[excid] * proportions[excid]
            for i in range(error_increase_interval - loop_count):
                alter_norms = []
                for cid in range(self.num_clients):
                    if np.any(np.isnan(sup_prop[cid])):
                        continue
                    alter_prop = del_prop + samples_per_client[excid] * sup_prop[cid]
                    alter_prop = alter_prop / alter_prop.sum()
                    error_alter = ((alter_prop - p) ** 2).sum()
                    alter_norms.append(error_alter)
                if min(alter_norms) < error_norm:
                    break
            if len(alter_norms) > 0 and min(alter_norms) < error_norm:
                alcid = np.argmin(alter_norms)
                proportions[excid] = sup_prop[alcid]
            loop_count += 1
        local_datas = [[] for _ in range(self.num_clients)]
        self.dirichlet_dist = []  # for efficiently visualizing
        for lb in lb_names:
            lb_idxs = lb_dict[lb]
            lb_proportion = np.array([pi[lb_names.index(lb)] * si for pi, si in zip(proportions, samples_per_client)])
            lb_proportion = lb_proportion / lb_proportion.sum()
            lb_proportion = (np.cumsum(lb_proportion) * len(lb_idxs)).astype(int)[:-1]
            lb_datas = np.split(lb_idxs, lb_proportion)
            self.dirichlet_dist.append([len(lb_data) for lb_data in lb_datas])
            local_datas = [local_data + lb_data.tolist() for local_data, lb_data in zip(local_datas, lb_datas)]
        self.dirichlet_dist = np.array(self.dirichlet_dist).T
        for i in range(self.num_clients): np.random.shuffle(local_datas[i])
        len_dist = [len(d) for d in local_datas]
        while min(len_dist)<=self.minvol:
            min_did = np.argmin(len_dist)
            max_did = np.argmax(len_dist)
            max_d = local_datas[max_did]
            min_d = local_datas[min_did]
            if len(max_d)<=self.minvol:
                raise RuntimeError("The number of clients is too large to distribute enough samples to each client when minvol=={}. Please decrease the number of clients".format(self.minvol))
            min_d.extend(max_d[:1])
            max_d = max_d[1:]
            local_datas[min_did] = min_d
            local_datas[max_did] = max_d
            len_dist = [len(d) for d in local_datas]
        self.local_datas = local_datas
        return local_datas

class DiversityPartitioner(BasicPartitioner):
    """`Partition the indices of samples in the original dataset according to numbers of types of a particular
    attribute (e.g. label) . This way of partition is widely used by existing works in federated learning.

    Args:
        num_clients (int, optional): the number of clients
        diversity (float, optional): the ratio of locally owned types of the attributes (i.e. the actual number=diversity * total_num_of_types)
        imbalance (float, optional): the degree of imbalance of the amounts of different local_movielens_recommendation data (0<=imbalance<=1)
        index_func (int, optional): the index of the distribution-dependent (i.e. label) attribute in each sample.
    """
    def __init__(self, num_clients=100, diversity=1.0, index_func=lambda X:[xi[-1] for xi in X]):
        self.num_clients = num_clients
        self.diversity = diversity
        self.index_func = index_func

    def __str__(self):
        name = "div{:.1f}".format(self.diversity)
        return name

    def __call__(self, data):
        labels = self.index_func(data)
        num_classes = len(set(labels))
        dpairs = [[did, lb] for did, lb in zip(list(range(len(data))), labels)]
        num = max(int(self.diversity * num_classes), 1)
        K = num_classes
        local_datas = [[] for _ in range(self.num_clients)]
        if num == K:
            for k in range(K):
                idx_k = [p[0] for p in dpairs if p[1] == k]
                np.random.shuffle(idx_k)
                split = np.array_split(idx_k, self.num_clients)
                for cid in range(self.num_clients):
                    local_datas[cid].extend(split[cid].tolist())
        else:
            times = [0 for _ in range(num_classes)]
            contain = []
            for i in range(self.num_clients):
                current = []
                j = 0
                while (j < num):
                    mintime = np.min(times)
                    ind = np.random.choice(np.where(times == mintime)[0])
                    if (ind not in current):
                        j = j + 1
                        current.append(ind)
                        times[ind] += 1
                contain.append(current)
            for k in range(K):
                idx_k = [p[0] for p in dpairs if p[1] == k]
                np.random.shuffle(idx_k)
                split = np.array_split(idx_k, times[k])
                ids = 0
                for cid in range(self.num_clients):
                    if k in contain[cid]:
                        local_datas[cid].extend(split[ids].tolist())
                        ids += 1
        return local_datas

class GaussianPerturbationPartitioner(BasicPartitioner):
    """`Partition the indices of samples I.I.D. and bind additional and static gaussian noise to each sample, which is
    a setting of feature skew in federated learning.

    Args:
        num_clients (int, optional): the number of clients
        imbalance (float, optional): the degree of imbalance of the amounts of different local_movielens_recommendation data (0<=imbalance<=1)
        sigma (float, optional): the degree of feature skew
        scale (float, optional): the standard deviation of noise
        index_func (int, optional): the index of the feature to be processed for each sample.
    """
    def __init__(self, num_clients=100, imbalance=0.0, sigma=0.1, scale=0.1, index_func=lambda X:[xi[0] for xi in X]):
        self.num_clients = num_clients
        self.imbalance = imbalance
        self.sigma = sigma
        self.scale = scale
        self.index_func = index_func

    def __str__(self):
        name = "perturb_gs{:.1f}_{:.1f}".format(self.sigma, self.scale)
        if self.imbalance > 0: name += '_imb{:.1f}'.format(self.imbalance)
        return name

    def __call__(self, data):
        shape = tuple(np.array(self.index_func(data)[0].shape))
        samples_per_client = self.data_imbalance_generator(self.num_clients, len(data), self.imbalance)
        d_idxs = np.random.permutation(len(data))
        local_datas = np.split(d_idxs, np.cumsum(samples_per_client))[:-1]
        local_datas = [di.tolist() for di in local_datas]
        local_perturbation_means = [np.random.normal(0, self.sigma, shape) for _ in range(self.num_clients)]
        local_perturbation_stds = [self.scale * np.ones(shape) for _ in range(self.num_clients)]
        local_perturbation = []
        for cid in range(self.num_clients):
            c_perturbation = [np.random.normal(local_perturbation_means[cid], local_perturbation_stds[cid]).tolist() for
                              _ in range(len(local_datas[cid]))]
            local_perturbation.append(c_perturbation)
        self.local_perturbation = local_perturbation
        return local_datas

class IDPartitioner(BasicPartitioner):
    """`Partition the indices of samples I.I.D. according to the ID of each sample, which requires the passed parameter
    `data` has attribution `id`.

    Args:
        num_clients (int, optional): the number of clients
        priority (str, optional): The value should be in set ('random', 'max', 'min'). If the number of clients is smaller than the total number of all the clients, this term will decide the selected clients according to their local_movielens_recommendation data sizes.
    """
    def __init__(self, num_clients=-1, priority='random', index_func=lambda X:X.id):
        self.num_clients = int(num_clients)
        self.priorty = priority
        self.index_func = index_func

    def __str__(self):
        return 'id'

    def __call__(self, data):
        all_data = list(range(len(data)))
        data_owners = self.index_func(data)
        local_datas = collections.defaultdict(list)
        for idx in range(len(all_data)):
            local_datas[data_owners[idx]].append(idx)
        local_datas = list(local_datas.values())
        if self.num_clients < 0:
            self.num_clients = len(local_datas)
        elif self.priorty == 'max':
            local_datas = sorted(local_datas, key=lambda x: len(x), reverse=True)[:self.num_clients]
        elif self.priorty == 'min':
            local_datas = sorted(local_datas, key=lambda x: len(x))[:self.num_clients]
        elif self.priorty == 'random':
            random.shuffle(local_datas)
            local_datas = local_datas[:self.num_clients]
        # local_datas = sorted(local_datas, key=lambda x:data[x[0]][] if self.index_func is not None else data.id[x[0]])
        return local_datas

class VerticalSplittedPartitioner(BasicPartitioner):
    """`Partition the indices and shapes of samples in the original dataset for vertical federated learning. Different
    to the above partitioners, the partitioner.__call__ returns more flexible partition information instead of the indices that
    can be used to rebuild the partitioned data.

    Args:
        num_parties (int, optional): the number of parties
        imbalance (float, optional): the degree of imbalance of the number of features
        dim (int, optional): the dim of features to be partitioned
    """
    def __init__(self, num_parties=-1, imbalance=0, dim=-1):
        self.num_parties = int(num_parties)
        self.imbalance = imbalance
        self.feature_pointers = []
        self.dim = dim

    def __str__(self):
        return 'vertical_splitted_IBM{}'.format(self.imbalance)

    def __call__(self, data):
        local_datas = []
        feature = data[0][0]
        shape = feature.shape
        if self.dim == -1: self.dim = int(np.argmax(shape))
        self.num_parties = min(shape[self.dim], self.num_parties)
        feature_sizes = self.gen_feature_size(shape[self.dim], self.num_parties, self.imbalance)
        for pid in range(self.num_parties):
            pdata = {'sample_idxs': list(range(len(data))), 'pt_feature': (self.dim, feature_sizes, pid),
                     'with_label': (pid == 0)}
            local_datas.append(pdata)
        return local_datas

    def gen_feature_size(self, total_size, num_parties, imbalance=0):
        size_partitions = []
        size_gen = self.integer_k_partition(total_size, num_parties)
        while True:
            try:
                tmp = next(size_gen)
                if tmp is not None:
                    size_partitions.append(tmp)
            except StopIteration:
                break
        size_partitions = sorted(size_partitions, key=lambda x: np.std(x))
        res = size_partitions[int(imbalance * (len(size_partitions) - 1))]
        return res

    def integer_k_partition(self, n, k, l=1):
        '''n is the integer to partition, k is the length of partitions, l is the min partition element size'''
        if k < 1:
            return None
        if k == 1:
            if n >= l:
                yield (n,)
            return None
        for i in range(l, n + 1):
            for result in self.integer_k_partition(n - i, k - 1, i):
                yield (i,) + result

class NodeLouvainPartitioner(BasicPartitioner):
    """
    Partition a graph into several subgraph by louvain algorithms. The input
    of this partitioner should be of type networkx.Graph
    """
    def __init__(self, num_clients=100):
        self.num_clients = num_clients

    def __str__(self):
        name = "Louvain"
        return name

    def __call__(self, data):
        r"""
        Partition graph data by Louvain algorithm and similar nodes (i.e. being of the same community) will be
        allocated one client.
        Args:
            data (networkx.Graph):
        Returns:
            local_nodes (List): the local nodes id owned by each client (e.g. [[1,2], [3,4]])
        """
        local_nodes = [[] for _ in range(self.num_clients)]
        self.node_groups = community.community_louvain.best_partition(data)
        groups = collections.defaultdict(list)
        for ni, gi in self.node_groups.items():
            groups[gi].append(ni)
        groups = {k: groups[k] for k in list(range(len(groups)))}
        # ensure the number of groups is larger than the number of clients
        while len(groups) < self.num_clients:
            # find the group with the largest size
            groups_lens = [groups[k] for k in range(len(groups))]
            max_gi = np.argmax(groups_lens)
            # set the size of the new group
            min_glen = min(groups_lens)
            max_glen = max(groups_lens)
            if max_glen < 2 * min_glen: min_glen = max_glen // 2
            # split the group with the largest size into two groups
            nodes_in_gi = groups[max_gi]
            new_group_id = len(groups)
            groups[new_group_id] = nodes_in_gi[:min_glen]
            groups[max_gi] = nodes_in_gi[min_glen:]
        # allocate different groups to clients
        groups_lens = [len(groups[k]) for k in range(len(groups))]
        group_ids = np.argsort(groups_lens)
        for gi in group_ids:
            cid = np.argmin([len(li) for li in local_nodes])
            local_nodes[cid].extend(groups[gi])
        return local_nodes

class BasicHierPartitioner(BasicPartitioner):
    def __init__(self, Partitioner1=IIDPartitioner, pargs1:dict ={'num_clients':5}, Partitioner2=IIDPartitioner, pargs2:dict = {'num_clients':20}):
        self.p1 = Partitioner1(**pargs1)
        self.p2 = Partitioner2(**pargs2)

    def __call__(self, data):
        edge_servers_data = self.p1(data)
        res = []
        for edge_data_idx in edge_servers_data:
            edge_data = [data[did] for did in edge_data_idx]
            edge_local_datas = self.p2(edge_data)
            for cid in range(len(edge_local_datas)):
                for k in range(len(edge_local_datas[cid])):
                    edge_local_datas[cid][k] = edge_data_idx[edge_local_datas[cid][k]]
            res.append(edge_local_datas)
        return res

class LinkIIDPartitioner(IIDPartitioner):
    """
        Partition a graph into several subgraph where there is no overlapping on edges for any two subgraphs. The input
        of this partitioner should be of type networkx.Graph
        """

    def __init__(self, num_clients=100, imbalance=0):
        super(LinkIIDPartitioner, self).__init__(num_clients, imbalance)

    def __str__(self):
        name = "link_iid"
        return name

    def __call__(self, data: nx.Graph):
        r"""
        Partition graph data by Louvain algorithm and similar nodes (i.e. being of the same community) will be
        allocated one client.
        Args:
            data (networkx.Graph):
        Returns:
            local_nodes (List): the local nodes id owned by each client (e.g. [[1,2], [3,4]])
        """
        data = list(data.edges)
        return super().__call__(data)

class LinkDirichletPartitioner(DirichletPartitioner):
    def __init__(self, num_clients=100, alpha=1.0, error_bar=1e-6, imbalance=0, index_func=lambda X:[xi[-1] for xi in X]):
        super(LinkDirichletPartitioner, self).__init__(num_clients=num_clients, alpha=alpha, error_bar=error_bar, imbalance=imbalance, index_func=index_func)

    def __call__(self, data):
        attrs = self.index_func(data)
        data = list(data.edges)
        num_attrs = len(set(attrs))
        samples_per_client = self.data_imbalance_generator(self.num_clients, len(data), self.imbalance)
        # count the label distribution
        lb_counter = collections.Counter(attrs)
        lb_names = list(lb_counter.keys())
        p = np.array([1.0 * v / len(data) for v in lb_counter.values()])
        lb_dict = {}
        attrs = np.array(attrs)
        for lb in lb_names:
            lb_dict[lb] = np.where(attrs == lb)[0]
        proportions = [np.random.dirichlet(self.alpha * p) for _ in range(self.num_clients)]
        while np.any(np.isnan(proportions)):
            proportions = [np.random.dirichlet(self.alpha * p) for _ in range(self.num_clients)]
        sorted_cid_map = {k: i for k, i in zip(np.argsort(samples_per_client), [_ for _ in range(self.num_clients)])}
        error_increase_interval = 500
        max_error = self.error_bar
        loop_count = 0
        crt_id = 0
        crt_error = 100000
        while True:
            if loop_count >= error_increase_interval:
                loop_count = 0
                max_error = max_error * 10
            # generate dirichlet distribution till ||E(proportion) - P(D)||<=1e-5*self.num_classes
            mean_prop = np.sum([pi * di for pi, di in zip(proportions, samples_per_client)], axis=0)
            mean_prop = mean_prop / mean_prop.sum()
            error_norm = ((mean_prop - p) ** 2).sum()
            if crt_error - error_norm >= max_error:
                print("Error: {:.8f}".format(error_norm))
                crt_error = error_norm
            if error_norm <= max_error:
                break
            excid = sorted_cid_map[crt_id]
            crt_id = (crt_id + 1) % self.num_clients
            sup_prop = [np.random.dirichlet(self.alpha * p) for _ in range(self.num_clients)]
            del_prop = np.sum([pi * di for pi, di in zip(proportions, samples_per_client)], axis=0)
            del_prop -= samples_per_client[excid] * proportions[excid]
            for i in range(error_increase_interval - loop_count):
                alter_norms = []
                for cid in range(self.num_clients):
                    if np.any(np.isnan(sup_prop[cid])):
                        continue
                    alter_prop = del_prop + samples_per_client[excid] * sup_prop[cid]
                    alter_prop = alter_prop / alter_prop.sum()
                    error_alter = ((alter_prop - p) ** 2).sum()
                    alter_norms.append(error_alter)
                if min(alter_norms) < error_norm:
                    break
            if len(alter_norms) > 0 and min(alter_norms) < error_norm:
                alcid = np.argmin(alter_norms)
                proportions[excid] = sup_prop[alcid]
            loop_count += 1
        local_datas = [[] for _ in range(self.num_clients)]
        self.dirichlet_dist = []  # for efficiently visualizing
        for lb in lb_names:
            lb_idxs = lb_dict[lb]
            lb_proportion = np.array([pi[lb_names.index(lb)] * si for pi, si in zip(proportions, samples_per_client)])
            lb_proportion = lb_proportion / lb_proportion.sum()
            lb_proportion = (np.cumsum(lb_proportion) * len(lb_idxs)).astype(int)[:-1]
            lb_datas = np.split(lb_idxs, lb_proportion)
            self.dirichlet_dist.append([len(lb_data) for lb_data in lb_datas])
            local_datas = [local_data + lb_data.tolist() for local_data, lb_data in zip(local_datas, lb_datas)]
        self.dirichlet_dist = np.array(self.dirichlet_dist).T
        for i in range(self.num_clients): np.random.shuffle(local_datas[i])
        self.local_datas = local_datas
        return local_datas

class DeconcatPartitioner(BasicPartitioner):
    def __init__(self, num_clients:int=-1, min_vol:int=3):
        self.num_clients = num_clients
        self.min_vol = min_vol

    def __str__(self):
        return 'Deconcat'

    def __call__(self, data):
        assert isinstance(data, ConcatDataset)
        crt_idx = 0
        local_datas = []
        for cdata in data.datasets:
            if len(cdata)<self.min_vol: continue
            local_datas.append([crt_idx+i for i in range(len(cdata))])
            crt_idx += len(cdata)
        if self.num_clients<0: self.num_clients = len(local_datas)
        self.num_clients = min(self.num_clients, len(local_datas))
        if self.num_clients<len(local_datas): local_datas = local_datas[:self.num_clients]
        return local_datas

class MultiConcatPartitioner(BasicPartitioner):
    def __init__(self, num_clients=-1, imbalance=0.0):
        self.num_clients = num_clients
        self.imbalance = imbalance

    def __str__(self):
        return 'MultiConcat'

    def __call__(self, data):
        assert isinstance(data, ConcatDataset)
        if self.num_clients==-1: self.num_clients = len(data.datasets)
        self.num_clients = min(self.num_clients, len(data.datasets))
        samples_per_client = self.data_imbalance_generator(self.num_clients, len(data.datasets), self.imbalance)
        d_idxs = np.random.permutation(len(data.datasets))
        local_datas = np.split(d_idxs, np.cumsum(samples_per_client))[:-1]
        local_datas = [di.tolist() for di in local_datas]
        return local_datas

# class KMeansPartitioner(BasicPartitioner):
#     """`Partition the indices of samples in the original dataset
#         according to their feature similarity.
#
#         Args:
#             num_clients (int, optional): the number of clients
#             imbalance (float, optional): the degree of imbalance of the amounts of different local_movielens_recommendation data (0<=imbalance<=1)
#         """
#
#     def __init__(self, num_clients=100, index_func=lambda x:x[0]):
#         self.num_clients = num_clients
#         self.index_func = index_func
#
#     def __str__(self):
#         name = ""
#         return name
#
#     def __call__(self, data):
#         features = [self.index_func(d) for d in data]
#         if len(features)==0: raise NotImplementedError("Feature indexed by {} was not found.".format(self.index_func))
#         if torch.is_tensor(features[0]):
#             features = np.array([t.view(-1).cpu().numpy() for t in features])
#         else:
#             features = np.array(features)
#         if len(features[0].shape)>1:
#             features = np.array([fe.flatten() for fe in features])
#         from scipy.cluster.vq import kmeans
#         centers = kmeans(features, self.num_clients)
#         local_datas = [[] for _ in self.num_clients]
#         for did, fe in enumerate(features):
#             cid = int((centers@fe).argmax())
#             local_datas[cid].append(did)
#         # local_datas = [di.tolist() for di in local_datas]
#         return local_datas


from torch import nn
from flgo.utils.fmodule import FModule

class Model(FModule):
    def __init__(self):
        super(Model, self).__init__()
        self.embedder = nn.Sequential(
            nn.Conv2d(3, 64, 5),
            nn.ReLU(),
            nn.MaxPool2d(2),
            nn.Conv2d(64, 64, 5),
            nn.ReLU(),
            nn.MaxPool2d(2),
            nn.Flatten(1),
            nn.Linear(1600, 384),
            nn.ReLU(),
            nn.Linear(384, 192),
            nn.ReLU(),
        )
        self.fc = nn.Linear(192, 10)

    def forward(self, x):
        x = self.get_embedding(x)
        return self.fc(x)

    def get_embedding(self, x):
        return self.embedder(x)

def init_local_module(object):
    pass

def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.model = Model().to(object.device)


import copy
import warnings

from flgo.simulator.base import BasicSimulator
import os
import zipfile
try:
    import pandas as pd
except:
    pd = None
import urllib
import numpy as np
import re
import random
"""
************************* Availability ***************************
We construct the client availability according to the public real-world dataset "Users Active 
Time Prediction". This is data dealing with mobile app usage of the customers, where an app 
has some personal information and online active timing of the customers. Whenever the 
customers login in-app and view anything, the app server gets pings from their mobile phone 
indicating that they are using the app.  More information about this dateset is in
https://www.kaggle.com/datasets/bhuvanchennoju/mobile-usage-time-prediction .
Before using this simulator, you should manually download and unzip the original file in kaggle to
ensure the  existence of file `benchmark/RAW_DATA/USER_ACTIVE_TIME/pings.csv`.
"""
class Simulator(BasicSimulator):
    def __init__(self, objects = [], option = None):
        super().__init__(objects)
        self.option = option
        self.num_clients = len(self.clients)
        self.customer_map = {}
        # availability
        self.init_availability(os.path.join('benchmark', 'RAW_DATA', 'USER_ACTIVE_TIME'))
        # computing resources
        self.init_computing_resource(os.path.join('benchmark', 'RAW_DATA', 'BODT'))
        # upload & download speed
        self.init_network(os.path.join('benchmark', 'RAW_DATA', 'OoklaSpeedtest'))

    def init_availability(self, rawdata_path):
        zipfile_path = os.path.join(rawdata_path,'archive.zip')
        data_path = os.path.join(rawdata_path,'pings.csv')
        if not os.path.exists(data_path) and os.path.exists(zipfile_path):
            f = zipfile.ZipFile(zipfile_path, 'r')
            for file in f.namelist():
                f.extract(file, rawdata_path)
        if not os.path.exists(data_path):
            raise FileNotFoundError("Please download the original dataset in https://www.kaggle.com/datasets/bhuvanchennoju/mobile-usage-time-prediction , and move it into `benchmark/RAW_DATA/USER_ACTIVE_TIME/pings.csv`")
        customers_info = pd.read_csv(os.path.join(rawdata_path, "customers.csv"))
        customers_info = customers_info.drop_duplicates(['id'], ignore_index=True)
        customers_availability = pd.read_csv(os.path.join(rawdata_path, 'pings.csv'))
        customers_availability['timestamp'] = customers_availability['timestamp'] - customers_availability['timestamp'][0]
        customers_availability_by_time = customers_availability.groupby('timestamp')['id'].apply(list)
        customers_availability_by_id = customers_availability.groupby('id')['timestamp'].apply(list)
        customers_availability_by_id = customers_availability_by_id.to_frame(name='timestamps')
        customers_availability_by_id = customers_availability_by_id.reset_index()
        if self.option['availability'] == 'IDL' or (self.option['availability']=='HIGH' and self.num_clients<len(customers_availability_by_id)) :
            customers_availability_by_id['num_stamps'] = customers_availability_by_id.apply(lambda x: len(x['timestamps']), axis=1)
            customers_availability_by_id = customers_availability_by_id.sort_values(by='num_stamps', ascending=False)
            customers_availability_by_id = customers_availability_by_id['id'].to_list()[:self.num_clients]
            self.customer_map = {cid: customers_availability_by_id[cid] for cid in range(self.num_clients)}
        elif self.option['availability'] == 'LOW' and self.num_clients<len(customers_availability_by_id):
            customers_availability_by_id['num_stamps'] = customers_availability_by_id.apply(lambda x: len(x['timestamps']), axis=1)
            customers_availability_by_id = customers_availability_by_id.sort_values(by='num_stamps', ascending=True)
            customers = customers_availability_by_id['id'].to_list()[:self.num_clients]
            self.customer_map = {cid: customers[cid] for cid in range(self.num_clients)}
        elif self.option['availability']=='RANDOM':
            replacement = True  if self.num_clients>len(customers_availability_by_id) else False
            customers = customers_availability_by_id.sample(n=self.num_clients, replace=replacement)['id'].to_list()
            self.customer_map = {cid: customers[cid] for cid in range(self.num_clients)}
        else:
            raise NotImplementedError("Availability {} has been not implemented.".format(self.option['availability']))
        self.availability_table = customers_availability_by_time
        # def visualize(a, b, k=10000):
        #     while a%15!=0: a=a+1
        #     import matplotlib.pyplot as plt
        #     ratio = max( (b-a)//k, 1)
        #     plt.figure(figsize=(ratio*2.56, 2.56))
        #     state = [0 for _ in range(len(customers_info))]
        #     time_start = [0 for _ in range(len(customers_info))]
        #     time_end = [0 for _ in range(len(customers_info))]
        #     for tid in range(a, b, 15):
        #         try:
        #             cids = customers_availability_by_time[tid]
        #         except:
        #             continue
        #         y = customers_info.loc[customers_info['id'].isin(cids)].index.to_list()
        #         for cid in range(len(customers_info)):
        #             if cid in y:
        #                 if state[cid]==0:
        #                     state[cid] = 1
        #                     time_start[cid] = tid
        #             else:
        #                 if state[cid]==1:
        #                     state[cid]=0
        #                     time_end[cid] = tid
        #                     plt.plot([time_start[cid], time_end[cid]], [cid, cid], c='g')
        #     plt.show()
        return

    def init_computing_resource(self, rawdata_path):
        # u_k = f(dev_k, model_k)
        # TimePerBatch_k =  u_k * batch_size * model_size
        # TimeLocalTraining_k = TimePerBatch * working_amount
        # k_dev should be initialized here
        github_url = "https://github.com/UbiquitousLearning/Benchmark-On-Device-Training/tree/master/tools/train/benchmark"
        device_names = ["MEIZU_Freq_1708800", "XIAOMI","MEIZU_Core_f0"]
        models = ["Lenet", "Squeezenet", "Mobilenet", "Alexnet"]
        model_sizes = {
            "Lenet": 3.2,
            "Squeezenet": 441.2,
            "Mobilenet": 3400,
            "Alexnet": 61000,
        }
        file_tmp = 'time_table_{}.csv'
        if not os.path.exists(os.path.join(rawdata_path, 'XIAOMI')):
            for dev in device_names:
                device_path = os.path.join(rawdata_path, dev)
                if not os.path.exists(device_path):
                    os.mkdir(device_path)
                for model in models:
                    model_url = os.path.join(github_url, dev, 'processed_data', model, file_tmp.format(model))
                    urllib.request.urlretrieve(model_url, device_path)
        us = []
        for dev in device_names:
            device_path = os.path.join(rawdata_path, dev)
            u_models = []
            for model in models:
                filename = 'time_table_{}.csv'.format(model)
                filepath = os.path.join(device_path, filename)
                if os.path.exists(filepath):
                    data = pd.read_csv(filepath)
                    batchsizes = data['batchsize'].to_list()
                    time_per_batch = data['expr_train'].to_list()
                    u_current_model = []
                    for bs, btime in zip(batchsizes, time_per_batch):
                        u_current_model.append(1.0*btime/model_sizes[model]/bs)
                    u_current_model = np.array(u_current_model).mean()
                    u_models.append(u_current_model)
            if len(u_models)>0: us.append(np.array(u_models).mean())
        us = np.array(us)
        if "LCLN" in self.option['responsiveness']:
            flag = self.option['responsiveness'].find('LCLN-')
            tmp  = self.option['responsiveness']
            if not tmp.endswith('_'): tmp = tmp + '_'
            std = float(tmp[flag+5: tmp.find('_', flag)])
            u_mean = us.mean()
            mu = np.log(u_mean) - std**2/2.0
            self.client_us = np.random.lognormal(mu, std, self.num_clients).tolist()
        else:
            ulow = us.min()
            uhigh = us.max()
            self.client_us = np.random.uniform(low = ulow, high = uhigh, size=self.num_clients).tolist()

    def init_network(self, rawdata_path):
        # load Ookla dataset
        zipfile_path = os.path.join(rawdata_path, 'archive.zip')
        data_path = os.path.join(rawdata_path, 'fixed_year_2020_quarter_01.csv')
        if not os.path.exists(data_path) and os.path.exists(zipfile_path):
            f = zipfile.ZipFile(zipfile_path, 'r')
            for file in f.namelist():
                f.extract(file, rawdata_path)
        if not os.path.exists(data_path):
            raise FileNotFoundError("Please download the original dataset in https://www.kaggle.com/datasets/dimitrisangelide/speedtest-data-by-ookla , and move it into `benchmark/RAW_DATA/OoklaSpeedtest/fixed_year_2020_quarter_01.csv`")

        data = pd.read_csv(data_path) # 获取colums
        # 数据集分为fixed和mobile两类，共22个csv文件，按年份和季度记录数据
        # 用到数据集中的Avg. Avg U Kbps（上传速度）和Avg. Avg D Kbps（下载速度）
        # 思路是把fixed和mobile的分开来然后合并所有的年份和季度数据，
        # 根据国家name去重（国家看作client），再将两个速度取均值
        Mobile_df = pd.DataFrame([], columns=data.columns)
        Broadband_df = pd.DataFrame([], columns=data.columns)
        for dirname, _, filenames in os.walk(rawdata_path):
            for filename in filenames:
                if filename.endswith('.zip'): continue
                meta_info = filename.split('/')[-1]
                data = pd.read_csv(dirname + '/' + filename, thousands=r',').convert_dtypes()
                if set(data.columns).intersection(set({'Number of Record': 'Number of Records'}.keys())):
                    data.rename(columns={'Number of Record': 'Number of Records'}, inplace=True)
                # data = self.col_name_corrections(data, {'Number of Record': 'Number of Records'})
                data['Year'] = np.int64(re.search('year_(.*)_quarter', meta_info).group(1))
                data['Quarter'] = np.int64(re.search('quarter_(.*).csv', meta_info).group(1))
                if 'mobile' in meta_info:
                    Mobile_df = pd.concat([Mobile_df, data])
                else:
                    Broadband_df = pd.concat([Broadband_df, data])

        Mobile_df = Mobile_df.astype({'Year': np.int64, 'Quarter': np.int64})
        Broadband_df = Broadband_df.astype({'Year': np.int64, 'Quarter': np.int64})
        Mobile_df.sort_values(by=['Year', 'Quarter'], ascending=[True, True], inplace=True)
        Broadband_df.sort_values(by=['Year', 'Quarter'], ascending=[True, True], inplace=True)
        unique_countries_broadband = Broadband_df.groupby('Name').agg({'Number of Records': 'sum',
                                                                       'Devices':'sum',
                                                                       'Tests':'sum',
                                                                       'Avg. Avg U Kbps': np.mean,
                                                                       'Avg. Avg D Kbps': np.mean,
                                                                       'Avg Lat Ms': np.mean,
                                                                       'Avg. Pop2005': np.mean,
                                                                       'Rank Upload': np.mean,
                                                                       'Rank Download': np.mean,
                                                                       'Rank Latency': np.mean,
                                                                       'Year': np.mean,
                                                                       'Quarter': np.mean,
                                                                       })
        unique_countries_mobile = Mobile_df.groupby('Name').agg({'Number of Records': 'sum',
                                                                       'Devices': 'sum',
                                                                       'Tests': 'sum',
                                                                       'Avg. Avg U Kbps': np.mean,
                                                                       'Avg. Avg D Kbps': np.mean,
                                                                       'Avg Lat Ms': np.mean,
                                                                       'Avg. Pop2005': np.mean,
                                                                       'Rank Upload': np.mean,
                                                                       'Rank Download': np.mean,
                                                                       'Rank Latency': np.mean,
                                                                       'Year': np.mean,
                                                                       'Quarter': np.mean,
                                                                       })
        # 分fixed和mobile根据上传，下载速度计算时间
        # 1 kbps = 10^3 bit/s
        broadband_id_list = [ i for i in range(unique_countries_broadband.shape[0])]
        broadband_upload_speed_list =  unique_countries_broadband['Avg. Avg U Kbps'].to_list()
        broadband_download_speed_list = unique_countries_broadband['Avg. Avg D Kbps'].to_list()
        # KB/s
        broadband_dev_list = [(u*125,d*125) for u,d in zip(broadband_upload_speed_list, broadband_download_speed_list)]
        mobile_id_list = [i for i in range(unique_countries_mobile.shape[0])]
        mobile_upload_speed_list = unique_countries_mobile['Avg. Avg U Kbps'].to_list()
        mobile_download_speed_list = unique_countries_mobile['Avg. Avg D Kbps'].to_list()
        mobile_dev_list = [(u*125,d*125) for u,d in zip(mobile_upload_speed_list, mobile_download_speed_list)]
        net_dev_list = broadband_dev_list + mobile_dev_list
        if 'NTH' in self.option['responsiveness'] or 'NTL' in self.option['responsiveness']:
            reverse = True if 'NTH' in self.option['responsiveness'] else False
            if self.num_clients <= len(net_dev_list):
                net_dev_list.sort(key = lambda x:x[0]+x[1], reverse=reverse)
                self.net_speeds = net_dev_list[:self.num_clients]
            else:
                self.net_speeds = net_dev_list * (self.num_clients/len(net_dev_list))
                num_rest = self.num_clients % len(net_dev_list)
                net_dev_list.sort(key=lambda x: x[0] + x[1], reverse=reverse)
                self.net_speeds = self.net_speeds + net_dev_list[:num_rest]
        else:
            replacement = True if self.num_clients > len(net_dev_list) else False
            net_ids = np.random.choice(list(range(len(net_dev_list))), size=self.num_clients, replace=replacement)
            self.net_speeds = [net_dev_list[nid] for nid in net_ids]

    def update_client_availability(self):
        t = self.gv.clock.current_time
        t = t%self.availability_table.index[-1]
        aid = t-t%15
        try:
            available_customers = self.availability_table[aid]
        except:
            available_customers = []
        pa, pua = [], []
        for cid in self.all_clients:
            pai = 1.0 if self.customer_map[cid] in available_customers else 0.0
            pa.append(pai)
            pua.append(1.0-pai)
        self.set_variable(self.all_clients, 'prob_available', pa)
        self.set_variable(self.all_clients, 'prob_unavailable', pua)

    def update_client_responsiveness(self, client_ids, *args, **kwargs):
        # calculate time of uploading and downloading model
        latency = []
        up_pkg_sizes = self.get_variable(client_ids, '__upload_package_size')
        down_pkg_sizes = self.get_variable(client_ids, '__download_package_size')
        for cid, upsize,down_size in zip(client_ids, up_pkg_sizes, down_pkg_sizes):
            latency.append(int(1.0*upsize/self.net_speeds[cid][0] + 1.0*down_size/self.net_speeds[cid][1]))
        # calculate time of local_movielens_recommendation computing
        tlcs = []
        model_sizes =  self.get_variable(client_ids, '__model_size')
        working_amounts = self.get_variable(client_ids, 'working_amount')
        for cid, model_size, wa in zip(client_ids, model_sizes, working_amounts):
            uk = self.client_us[cid]
            tlc_k = uk*model_size/1000.0*self.clients[cid].batch_size*wa/1000.0
            tlcs.append(tlc_k)
        for idx, _ in enumerate(client_ids):
            latency[idx] = int(latency[idx] + tlcs[idx])
        self.set_variable(client_ids, 'latency', latency)

    def update_client_completeness(self, client_ids, *args, **kwargs):
        pass


import flgo.benchmark.toolkits.partition
from .model import mf
default_model = mf
default_partitioner = flgo.benchmark.toolkits.partition.IDPartitioner
default_partition_para = {'index_func':lambda x:x[0]}



from torch import nn
import torch.nn.functional as F

import flgo.benchmark
from flgo.utils.fmodule import FModule

import os
from flgo.benchmark.toolkits.series.classification.datasets import UCRArchiveDataset

class Model(FModule):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv1d(in_channels=1, out_channels=32, kernel_size=3, padding=1)
        self.conv2 = nn.Conv1d(in_channels=32, out_channels=64, kernel_size=3, padding=1)
        self.fc1 = nn.Linear(704, 512)
        self.fc2 = nn.Linear(512, 24)

    def forward(self, x):
        x = self.get_embedding(x)
        x = self.fc2(x)
        return x

    def get_embedding(self, x):
        if len(x.shape) == 1:
            x = x.unsqueeze(0)
        x = x.view(x.shape[0], 1, 46)
        x = self.conv1(x)
        x = F.relu(x)
        x = F.max_pool1d(x, 2)
        x = self.conv2(x)
        x = F.relu(x)
        x = F.max_pool1d(x, 2)
        x = x.view(-1, x.shape[1]*x.shape[2])
        x = F.relu(self.fc1(x))
        return x


def init_local_module(object):
    pass


def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.model = Model().to(object.device)



from flgo.benchmark.base import BasicTaskGenerator, BasicTaskCalculator
from flgo.benchmark.base import XHorizontalTaskPipe as TaskPipe
import numpy as np
import torch
from torch.utils.data import DataLoader

class TaskGenerator(BasicTaskGenerator):
    def __init__(self, dimension=2, num_clients = 30, data_size = 400, std=1.0):
        super(TaskGenerator, self).__init__('distributed_quadratic_programming', '')
        self.dimension = dimension
        self.num_clients = num_clients
        self.data_size = data_size
        self.std = std

    def load_data(self):
        es = [[] for _ in range(self.num_clients)]
        emeans = []
        for i in range(self.num_clients):
            mean_ei = np.random.normal(0, self.std, self.dimension)
            emeans.append(mean_ei)
            es[i].extend(np.random.multivariate_normal(mean_ei, 0.05*np.eye(self.dimension), self.data_size).tolist())
        mean_e = np.mean(emeans, axis=0)
        self.test_data = {'x': [mean_e.tolist()]}
        self.local_datas = []
        for i in range(self.num_clients):
            self.local_datas.append({'x':es[i]})

    def get_task_name(self):
        return 'distqp_{}'.format(self.num_clients)

class TaskCalculator(BasicTaskCalculator):
    def __init__(self, device, optimizer_name='sgd'):
        super(TaskCalculator, self).__init__(device, optimizer_name)
        self.DataLoader = DataLoader

    def compute_loss(self, model, data):
        tdata = self.to_device(data)
        outputs = model(tdata)
        loss = 0.5 * torch.mean(outputs)
        return {'loss': loss}

    @torch.no_grad()
    def test(self, model, data, batch_size=1, num_workers=0, pin_memory=False):
        tdata = self.to_device(data[0])
        outputs = model(tdata)
        loss = 0.5 * torch.mean(outputs)
        return {'loss': loss.item()}

    def to_device(self, data):
        return data[0].to(self.device)

    def get_dataloader(self, dataset, batch_size=64, shuffle=True, num_workers=0, pin_memory=False, drop_last=False, *args, **kwargs):
        if self.DataLoader == None:
            raise NotImplementedError("DataLoader Not Found.")
        return DataLoader(dataset, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers)


import os
import math
import torch
import torch.utils.data
from flgo.benchmark.toolkits.nlp.classification import GeneralCalculator
from flgo.benchmark.base import FromDatasetPipe, FromDatasetGenerator, BasicTaskCalculator
from torchtext.data.functional import to_map_style_dataset
from torch.nn.utils.rnn import pad_sequence
try:
    import ujson as json
except:
    import json
from .config import train_data
try:
    from .config import padding_value
except:
    padding_value = 0
try:
    from .config import test_data
except:
    test_data = None
try:
    from .config import val_data
except:
    val_data = None

def collate_fn(batch):
    src_batch, tgt_batch = [], []
    for src_sample, tgt_sample in batch:
        if isinstance(src_sample, list):
            src_batch.append(torch.tensor(src_sample, dtype=torch.int64))
            tgt_batch.append(torch.tensor(tgt_sample, dtype=torch.int64))

        else:
            src_batch.append(src_sample)
            tgt_batch.append(tgt_sample)
    src_batch = pad_sequence(src_batch, padding_value=padding_value)
    tgt_batch = pad_sequence(tgt_batch, padding_value=padding_value)
    return src_batch, tgt_batch

class TaskGenerator(FromDatasetGenerator):
    def __init__(self):
        super(TaskGenerator, self).__init__(benchmark=os.path.split(os.path.dirname(__file__))[-1],
                                            train_data=train_data, val_data=val_data, test_data=test_data)

    def prepare_data_for_partition(self):
        return to_map_style_dataset(self.train_data)

class TaskPipe(FromDatasetPipe):
    TaskDataset = torch.utils.data.Subset
    def __init__(self, task_path):
        super(TaskPipe, self).__init__(task_path, train_data=train_data, val_data=val_data, test_data=test_data)

    def save_task(self, generator):
        client_names = self.gen_client_names(len(generator.local_datas))
        feddata = {'client_names': client_names,}
        for cid in range(len(client_names)): feddata[client_names[cid]] = {'data': generator.local_datas[cid],}
        with open(os.path.join(self.task_path, 'data.json'), 'w') as outf:
            json.dump(feddata, outf)
        return

    def load_data(self, running_time_option) -> dict:
        if self.test_data is not None:
            test_data = to_map_style_dataset(self.test_data)
            if self.val_data is not None:
                server_test_data = test_data
                server_val_data = to_map_style_dataset(self.val_data)
            elif running_time_option['test_holdout'] > 0:
                server_test_data, server_val_data = self.split_dataset(test_data, running_time_option['test_holdout'])
            else:
                server_test_data = test_data
                server_val_data = None
        elif self.val_data is not None:
            server_test_data = None
            server_val_data = to_map_style_dataset(self.val_data)
        else:
            server_test_data = server_val_data = None
        # rearrange data for server
        task_data = {'server': {'test': server_test_data, 'val': server_val_data}}
        train_data = to_map_style_dataset(self.train_data)
        for cid, cname in enumerate(self.feddata['client_names']):
            cdata = self.TaskDataset(train_data, self.feddata[cname]['data'])
            if running_time_option['train_holdout'] > 0:
                cdata_train, cdata_val = self.split_dataset(cdata, running_time_option['train_holdout'])
                if running_time_option['local_test']:
                    cdata_val, cdata_test = self.split_dataset(cdata_val, running_time_option['local_test_ratio'])
                else:
                    cdata_test = None
            else:
                cdata_train = cdata
                cdata_val, cdata_test = None, None
            task_data[cname] = {'train': cdata_train, 'val': cdata_val, 'test': cdata_test}
        return task_data

class TaskCalculator(BasicTaskCalculator):
    def __init__(self, device, optimizer_name='sgd'):
        super(TaskCalculator, self).__init__(device, optimizer_name)
        self.DataLoader = torch.utils.data.DataLoader
        self.collate_fn = collate_fn
        self.criterion = self.loss_func

    def loss_func(self, outputs, targets, ignore_index=-100):
        loss_func = torch.nn.CrossEntropyLoss(ignore_index=ignore_index)
        return loss_func(outputs[1:].view(-1, outputs.shape[-1]), targets[1:].view(-1))

    def compute_loss(self, model, data):
        """
        Args:
            model: the model to train
            data: the training dataset
        Returns: dict of train-one-step's result, which should at least contains the key 'loss'
        """
        sources, targets = self.to_device(data)
        outputs = model(sources, targets)
        loss = self.criterion(outputs, targets, padding_value)
        return {'loss': loss}

    @torch.no_grad()
    def test(self, model, dataset, batch_size=64, num_workers=0, pin_memory=False):
        """
        Metric = [mean_accuracy, mean_loss]

        Args:
            model:
            dataset:
            batch_size:
        Returns: [mean_accuracy, mean_loss]
        """
        model.eval()
        if batch_size==-1:batch_size=len(dataset)
        data_loader = self.get_dataloader(dataset, batch_size=batch_size, num_workers=num_workers, pin_memory=pin_memory)
        total_loss = 0.0
        for batch_id, batch_data in enumerate(data_loader):
            batch_data = self.to_device(batch_data)
            outputs = model(batch_data[0], batch_data[1])
            batch_mean_loss = self.criterion(outputs, batch_data[1], model.ignore_index if hasattr(model, 'ignore_index') else -100).item()
            total_loss += batch_mean_loss * batch_data[-1].shape[-1]
        total_loss = total_loss/len(dataset)
        ppl = math.exp(total_loss)
        return {'loss':total_loss, 'ppl':ppl}

    def to_device(self, data):
        return data[0].to(self.device), data[1].to(self.device)

    def get_dataloader(self, dataset, batch_size=64, shuffle=True, num_workers=0, pin_memory=False, drop_last=False, *args, **kwargs):
        if self.DataLoader == None:
            raise NotImplementedError("DataLoader Not Found.")
        return self.DataLoader(dataset, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers, pin_memory=pin_memory, drop_last=drop_last, collate_fn=self.collate_fn)



from flgo.experiment.logger import BasicLogger

class ParallelHFLLogger(BasicLogger):
    def log_once(self, *args, **kwargs):
        test_metrics = self.object.test()
        for met, val in test_metrics.items():
            self.output[self.object.name+'_test_'+met].append(val)
        self.show_current_output()


from flgo.benchmark.base import *
from flgo.benchmark.toolkits.cv.classification import GeneralCalculator
from flgo.benchmark.toolkits.partition import BasicPartitioner
import numpy as np

TaskPipe = XYHorizontalTaskPipe
TaskCalculator = GeneralCalculator

class TaskGenerator(BasicTaskGenerator):
    def __init__(self, *args, **kwargs):
        super().__init__('fcube_classification', '')
        self.num_clients = 4
        self.num_classes = 2

    def load_data(self):
        X_train, y_train = [], []
        for loc in range(self.num_clients):
            for i in range(1000):
                p1 = random.random()
                p2 = random.random()
                p3 = random.random()
                if loc > 1:
                    p2 = -p2
                if loc % 2 == 1:
                    p3 = -p3
                if i % 2 == 0:
                    X_train.append([p1, p2, p3])
                    y_train.append(0)
                else:
                    X_train.append([-p1, -p2, -p3])
                    y_train.append(1)
        X_test, y_test = [], []
        for i in range(1000):
            p1 = random.random() * 2 - 1
            p2 = random.random() * 2 - 1
            p3 = random.random() * 2 - 1
            X_test.append([p1, p2, p3])
            if p1 > 0:
                y_test.append(0)
            else:
                y_test.append(1)
        idxs = np.linspace(0, 3999, 4000, dtype=np.int64)
        batch_idxs = np.array_split(idxs, self.num_clients)
        self.local_datas = [{'x': [X_train[did] for did in batch], 'y':[y_train[did] for did in batch]} for batch in batch_idxs]
        self.test_data = {'x':X_test, 'y':y_test}
        return

    def get_task_name(self):
        partitioner_name = 'feature_skew'
        return '_'.join(['B-'+self.benchmark, 'P-'+partitioner_name, 'N-'+str(self.num_clients)])



import torch
import flgo.benchmark.toolkits.cv.segmentation.transforms as T


class SegmentationPresetTrain:
    def __init__(self, *, base_size, crop_size, hflip_prob=0.5, mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225)):
        min_size = int(0.5 * base_size)
        max_size = int(2.0 * base_size)

        trans = [T.RandomResize(min_size, max_size)]
        if hflip_prob > 0:
            trans.append(T.RandomHorizontalFlip(hflip_prob))
        trans.extend(
            [
                T.RandomCrop(crop_size),
                T.PILToTensor(),
                T.ConvertImageDtype(torch.float),
                T.Normalize(mean=mean, std=std),
            ]
        )
        self.transforms = T.Compose(trans)

    def __call__(self, img, target):
        return self.transforms(img, target)


class SegmentationPresetEval:
    def __init__(self, *, base_size, mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225)):
        self.transforms = T.Compose(
            [
                T.RandomResize(base_size, base_size),
                T.PILToTensor(),
                T.ConvertImageDtype(torch.float),
                T.Normalize(mean=mean, std=std),
            ]
        )

    def __call__(self, img, target):
        return self.transforms(img, target)


from torchtext.datasets import Multi30k
import flgo.benchmark
import os
from torchtext.data.utils import get_tokenizer
from torchtext.vocab import build_vocab_from_iterator
import torch.nn as nn
import torch
import torch.nn.functional as F
import random

# 0. 加载数据集
language_pair = ['de', 'en']
path = os.path.join(flgo.benchmark.data_root,  'MULTI30K')
train_data, val_data, test_data = Multi30k(split=('train', 'valid', 'test'), language_pair=language_pair)

# 1. 加载tokenizer和词表
# init tokenizers
tokenizers = {}
tokenizers[language_pair[0]] = get_tokenizer('spacy', language='de_core_news_sm')
tokenizers[language_pair[1]] = get_tokenizer('spacy', language='en_core_web_sm')

#  init vocabs
vocabs = {}
PAD_IDX, UNK_IDX, BOS_IDX, EOS_IDX = 0, 1, 2, 3
special_symbols = ['<pad>', '<unk>', '<bos>', '<eos>']
for i,ln in enumerate(language_pair):
    # Create torchtext's Vocab object
    tokenizer = tokenizers[ln]
    all_tokens = [tokenizer(d[i]) for d in train_data]
    vocabs[ln] = build_vocab_from_iterator(all_tokens,
                                           min_freq=1,
                                           specials=special_symbols,
                                           special_first=True)
for ln in language_pair: vocabs[ln].set_default_index(UNK_IDX)

# 2. 把数据集中的字符串(源语言和目标语言)，根据tokenizer和vocab转化为数值向量
def sequential_transforms(*transforms):
    def func(txt_input):
        for transform in transforms:
            txt_input = transform(txt_input)
        return txt_input
    return func

def tensor_transform(token_ids):
    return torch.cat((torch.tensor([BOS_IDX]),
                      torch.tensor(token_ids),
                      torch.tensor([EOS_IDX])))

text_transform = {}
for ln in language_pair:
    text_transform[ln] = sequential_transforms(tokenizers[ln],  #Tokenization
                                               vocabs[ln],  #Numericalization
                                               tensor_transform) # Add BOS/EOS and create tensor

def apply_transform(x):
    return text_transform[language_pair[0]](x[0].rstrip("\n")), text_transform[language_pair[1]](x[1].rstrip("\n"))

train_data = train_data.map(apply_transform)
val_data = val_data.map(apply_transform)
test_data = test_data.map(apply_transform)

# 3. 定义模型
def get_model():
    INPUT_DIM = len(vocabs[language_pair[0]])
    OUTPUT_DIM = len(vocabs[language_pair[1]])
    ENC_EMB_DIM = 256
    DEC_EMB_DIM = 256
    ENC_HID_DIM = 512
    DEC_HID_DIM = 512
    ENC_DROPOUT = 0.5
    DEC_DROPOUT = 0.5
    attn = Attention(ENC_HID_DIM, DEC_HID_DIM)
    enc = Encoder(INPUT_DIM, ENC_EMB_DIM, ENC_HID_DIM, DEC_HID_DIM, ENC_DROPOUT)
    dec = Decoder(OUTPUT_DIM, DEC_EMB_DIM, ENC_HID_DIM, DEC_HID_DIM, DEC_DROPOUT, attn)
    model = Seq2Seq(enc, dec)
    model.apply(init_weights)
    return model

def init_weights(m):
    for name, param in m.named_parameters():
        if 'weight' in name:
            nn.init.normal_(param.data, mean=0, std=0.01)
        else:
            nn.init.constant_(param.data, 0)

class Encoder(nn.Module):
    def __init__(self, input_dim, emb_dim, enc_hid_dim, dec_hid_dim, dropout):
        super().__init__()
        self.embedding = nn.Embedding(input_dim, emb_dim)
        self.rnn = nn.GRU(emb_dim, enc_hid_dim, bidirectional=True)
        self.fc = nn.Linear(enc_hid_dim * 2, dec_hid_dim)
        self.dropout = nn.Dropout(dropout)

    def forward(self, src):
        embedded = self.dropout(self.embedding(src))
        outputs, hidden = self.rnn(embedded)
        hidden = torch.tanh(self.fc(torch.cat((hidden[-2, :, :], hidden[-1, :, :]), dim=1)))
        return outputs, hidden

class Attention(nn.Module):
    def __init__(self, enc_hid_dim, dec_hid_dim):
        super().__init__()
        self.attn = nn.Linear((enc_hid_dim * 2) + dec_hid_dim, dec_hid_dim)
        self.v = nn.Linear(dec_hid_dim, 1, bias=False)

    def forward(self, hidden, encoder_outputs):
        batch_size = encoder_outputs.shape[1]
        src_len = encoder_outputs.shape[0]
        hidden = hidden.unsqueeze(1).repeat(1, src_len, 1)
        encoder_outputs = encoder_outputs.permute(1, 0, 2)
        energy = torch.tanh(self.attn(torch.cat((hidden, encoder_outputs), dim=2)))
        attention = self.v(energy).squeeze(2)
        return F.softmax(attention, dim=1)

class Decoder(nn.Module):
    def __init__(self, output_dim, emb_dim, enc_hid_dim, dec_hid_dim, dropout, attention):
        super().__init__()
        self.output_dim = output_dim
        self.attention = attention
        self.embedding = nn.Embedding(output_dim, emb_dim)
        self.rnn = nn.GRU((enc_hid_dim * 2) + emb_dim, dec_hid_dim)
        self.fc_out = nn.Linear((enc_hid_dim * 2) + dec_hid_dim + emb_dim, output_dim)
        self.dropout = nn.Dropout(dropout)

    def forward(self, input, hidden, encoder_outputs):
        input = input.unsqueeze(0)
        embedded = self.dropout(self.embedding(input))
        a = self.attention(hidden, encoder_outputs)
        a = a.unsqueeze(1)
        encoder_outputs = encoder_outputs.permute(1, 0, 2)
        weighted = torch.bmm(a, encoder_outputs)
        weighted = weighted.permute(1, 0, 2)
        rnn_input = torch.cat((embedded, weighted), dim=2)
        output, hidden = self.rnn(rnn_input, hidden.unsqueeze(0))
        assert (output == hidden).all()
        embedded = embedded.squeeze(0)
        output = output.squeeze(0)
        weighted = weighted.squeeze(0)
        prediction = self.fc_out(torch.cat((output, weighted, embedded), dim=1))
        return prediction, hidden.squeeze(0)

class Seq2Seq(nn.Module):
    def __init__(self, encoder, decoder):
        super().__init__()
        self.encoder = encoder
        self.decoder = decoder

    def forward(self, src, trg, teacher_forcing_ratio=0.5):
        dev = src.device
        batch_size = src.shape[1]
        trg_len = trg.shape[0]
        trg_vocab_size = self.decoder.output_dim
        outputs = torch.zeros(trg_len, batch_size, trg_vocab_size).to(dev)
        encoder_outputs, hidden = self.encoder(src)
        input = trg[0, :]
        for t in range(1, trg_len):
            output, hidden = self.decoder(input, hidden, encoder_outputs)
            outputs[t] = output
            teacher_force = random.random() < teacher_forcing_ratio
            top1 = output.argmax(1)
            input = trg[t] if teacher_force else top1
        return outputs